from __future__ import annotations

from datetime import datetime

from . import db

DEFAULT_VARCHAR_LENGTH = 255


class OAuth(db.Model):
    __tablename__ = "oauth"

    id = db.Column(db.Integer, primary_key=True)
    user_id = db.Column(db.Integer, db.ForeignKey("user.id"))
    provider_id = db.Column(db.String(DEFAULT_VARCHAR_LENGTH))
    provider_user_id = db.Column(db.Integer)
    access_token = db.Column(db.String(DEFAULT_VARCHAR_LENGTH))
    created_at = db.Column(db.DateTime())
    id_token = db.Column(db.String())

    user = db.relationship("User", backref="identities")

    @classmethod
    def from_token(cls, provider_id, provider_user_id, token, user=None):
        return cls(
            user=user,
            provider_id=provider_id,
            provider_user_id=provider_user_id,
            access_token=token["access_token"],
            created_at=datetime.utcfromtimestamp(token["created_at"]),
            id_token=token["id_token"],
        )



from __future__ import annotations

import asyncio
from typing import Iterator

import aiohttp
import sqlalchemy
import yaml

STATUS_OK = 200


async def fetch(url, session):
    async with session.get(url) as response:
        if response.status != STATUS_OK:
            response.raise_for_status()
        return await response.text()


async def get_futures(urls: Iterator[str], timeout=None, headers=None):
    tasks = []
    async with aiohttp.ClientSession(headers=headers, timeout=timeout) as session:
        for url in urls:
            task = asyncio.ensure_future(fetch(url, session))
            tasks.append(task)
        responses = await asyncio.gather(*tasks)
    return responses


def fetch_urls(
    urls: list, headers: dict = None, timeout: aiohttp.ClientTimeout = None
) -> Iterator[str]:
    """Asynchronously fetch multiple urls and return `Iterator` object that contains all of the responses.

    Examples:
        ```python
        urls = ['https://httpbin.org/get', 'https://google.com', 'https://gitlab.com/meltano/']
        for result in fetch_urls(urls):
            print(result)
        ```

    Args:
        headers: Dictionary of headers to be passed with the request.
        timeout: Timeout object to be passed to the aiohttp lib.
        urls: Iterator of url strings.
    """
    try:
        loop = asyncio.get_event_loop()
    except RuntimeError:
        # Fix for when asyncio runs inside a sub-thread so there is no event loop available
        asyncio.set_event_loop(asyncio.new_event_loop())
        loop = asyncio.get_event_loop()

    yield from loop.run_until_complete(
        asyncio.ensure_future(get_futures(urls=urls, headers=headers, timeout=timeout))
    )


SQLALCHEMY_COL_TYPE_NAME_MAP = {
    "timestamp without time zone": sqlalchemy.TIMESTAMP(timezone=False),
    "timestamp with time zone": sqlalchemy.TIMESTAMP(timezone=True),
    "date": sqlalchemy.Date,
    "real": sqlalchemy.REAL,
    "integer": sqlalchemy.Integer,
    "smallint": sqlalchemy.SMALLINT,
    "text": sqlalchemy.TEXT,
    "bigint": sqlalchemy.BIGINT,
    "float": sqlalchemy.Float,
    "boolean": sqlalchemy.Boolean,
    "json": sqlalchemy.JSON,
}


def get_sqlalchemy_col(
    column_name: str, column_type_name: str, primary_key_col_name: str
) -> sqlalchemy.Column:
    """Return the sqlalchemy `Column` object to be used for `Table` definition."""
    col = sqlalchemy.Column(
        column_name,
        SQLALCHEMY_COL_TYPE_NAME_MAP.get(column_type_name, sqlalchemy.String),
    )
    if column_name == primary_key_col_name:
        col.primary_key = True
    return col


def tables_from_manifest(
    manifest_file_path: str, metadata: sqlalchemy.MetaData, schema_name: str
) -> dict[str, sqlalchemy.Table]:
    with open(manifest_file_path) as file:
        schema_manifest: dict = yaml.safe_load(file)
        tables = {}
        for table_name, schema in schema_manifest.items():
            primary_key_col_name = schema["primary_key"]
            columns: dict[str, str] = schema["columns"]
            columns_list = [
                get_sqlalchemy_col(column_name, column_type_name, primary_key_col_name)
                for column_name, column_type_name in columns.items()
            ]
            tables[table_name] = sqlalchemy.Table(
                table_name, metadata, *columns_list, schema=schema_name
            )
    return tables



"""Plugin glue code for Superset."""
from __future__ import annotations

import logging
import subprocess

import structlog

from meltano.core.behavior.hookable import hook
from meltano.core.error import AsyncSubprocessError
from meltano.core.plugin.error import PluginExecutionError
from meltano.core.plugin_invoker import PluginInvoker
from meltano.core.setting_definition import SettingDefinition

from . import BasePlugin, PluginType

logger = structlog.getLogger(__name__)


class SupersetInvoker(PluginInvoker):
    """Invoker that prepares env for Superset."""

    def env(self):
        """Environment variables for Superset.

        Returns:
            Dictionary of environment variables.
        """
        env = super().env()

        env["SUPERSET_HOME"] = str(self.plugin_config_service.run_dir)
        env["SUPERSET_CONFIG_PATH"] = str(self.files["config"])
        env["FLASK_APP"] = "superset"

        return env


class Superset(BasePlugin):
    """Plugin glue code for Superset."""

    __plugin_type__ = PluginType.UTILITIES

    invoker_class = SupersetInvoker

    EXTRA_SETTINGS = [SettingDefinition(name="_config_path")]

    @property
    def config_files(self):
        """Return the configuration files required by the plugin.

        Returns:
            Dictionary of config file identifiers and filenames
        """
        return {"config": "superset_config.py"}

    @hook("before_configure")
    async def before_configure(self, invoker: SupersetInvoker, session):  # noqa: WPS217
        """Write plugin configuration to superset_config.py.

        Args:
            invoker: the active PluginInvoker
            session: metadata database session

        Raises:
            PluginExecutionError: if config file couldn't be found
        """
        config = invoker.plugin_config_processed

        config_script_lines = [
            "import sys",
            "module = sys.modules[__name__]",
            f"config = {config!r}",
            "for key, value in config.items():",
            "    if key.isupper():",
            "        setattr(module, key, value)",
        ]

        custom_config_filename = invoker.plugin_config_extras["_config_path"]
        if custom_config_filename:
            custom_config_path = invoker.project.root.joinpath(custom_config_filename)

            if custom_config_path.exists():
                config_script_lines.extend(
                    [
                        "from importlib.util import module_from_spec, spec_from_file_location",
                        f'spec = spec_from_file_location("superset_config", {str(custom_config_path)!r})',
                        "custom_config = module_from_spec(spec)",
                        'sys.modules["superset_config"] = custom_config',
                        "spec.loader.exec_module(custom_config)",
                        "for key in dir(custom_config):",
                        "    if key.isupper():",
                        "        setattr(module, key, getattr(custom_config, key))",
                    ]
                )

                logger.info(f"Merged in config from {custom_config_path}")
            else:
                raise PluginExecutionError(
                    f"Could not find config file {custom_config_path}"
                )

        config_path = invoker.files["config"]
        with open(config_path, "w") as config_file:
            config_file.write("\n".join(config_script_lines))
        logging.debug(f"Created configuration at {config_path}")

    @hook("before_invoke")
    async def db_upgrade_hook(self, invoker: PluginInvoker, exec_args: list[str]):
        """Create or upgrade metadata database.

        Args:
            invoker: the active PluginInvoker
            exec_args: the args being passed

        Raises:
            AsyncSubprocessError: if command failed to run
        """
        handle = await invoker.invoke_async(
            "db",
            "upgrade",
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
        )
        exit_code = await handle.wait()

        if exit_code:
            raise AsyncSubprocessError(
                "Superset metadata database could not be initialized: `superset db upgrade` failed",
                handle,
            )

        logging.debug("Completed `superset db upgrade`")

    @hook("before_invoke")
    async def init_hook(self, invoker: PluginInvoker, exec_args: list[str]):
        """Create default roles and permissions.

        Args:
            invoker: the active PluginInvoker
            exec_args: the args being passed

        Raises:
            AsyncSubprocessError: if command failed to run
        """
        handle = await invoker.invoke_async(
            "init",
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
        )
        exit_code = await handle.wait()

        if exit_code:
            raise AsyncSubprocessError(
                "Superset default roles and permissions could not be created: `superset init` failed",
                handle,
            )

        logging.debug("Completed `superset init`")

    @hook("before_cleanup")
    async def before_cleanup(self, invoker: PluginInvoker):
        """Delete the config file.

        Args:
            invoker: the active PluginInvoker
        """
        config_file = invoker.files["config"]
        try:
            config_file.unlink()
            logging.debug(f"Deleted configuration at {config_file}")
        except FileNotFoundError:
            pass



"""Validation command."""

from __future__ import annotations

import asyncio
import shutil
import sys
from typing import Iterable

import click
import structlog
from sqlalchemy.orm.session import sessionmaker

from meltano.cli.utils import InstrumentedCmd, propagate_stop_signals
from meltano.core.db import project_engine
from meltano.core.project import Project
from meltano.core.validation_service import ValidationOutcome, ValidationsRunner

from . import cli
from .params import pass_project

logger = structlog.getLogger(__name__)

TEST_LINE_LENGTH = 60


def write_sep_line(title: str, sepchar: str, **kwargs):
    """Write a separator line in the terminal."""
    terminal_width, _ = shutil.get_terminal_size()
    char_count = (
        terminal_width - len(title) - 2  # Whitespace between sepchar and title
    ) // (2 * len(sepchar))
    char_count = max(char_count, 1)

    fill = sepchar * char_count
    line = f"\n{fill} {title} {fill}"

    if len(line) + len(sepchar.rstrip()) <= terminal_width:
        line += sepchar.rstrip()

    click.secho(line, **kwargs)


class CommandLineRunner(ValidationsRunner):
    """Validator that runs in the CLI."""

    async def run_test(self, name: str) -> int:
        """Run a test command.

        Args:
            name: Test command name to invoke.

        Returns:
            Exit code for the plugin invocation.
        """
        write_sep_line(f"{self.plugin_name}:{name}", "=", bold=True)
        handle = await self.invoker.invoke_async(command=name)
        with propagate_stop_signals(handle):
            exit_code = await handle.wait()
        return exit_code


@cli.command(cls=InstrumentedCmd, short_help="Run validations using plugins' tests.")
@click.option(
    "--all",
    "all_tests",
    is_flag=True,
    help="Run all defined tests in all plugins.",
)
@click.argument(
    "plugin_tests",
    default=None,
    required=False,
    nargs=-1,
)
@pass_project(migrate=True)
def test(
    project: Project,
    all_tests: bool,
    plugin_tests: tuple[str] = (),
):
    """
    Run validations using plugins' tests.

    \b\nRead more at https://docs.meltano.com/reference/command-line-interface#test
    """
    _, session_maker = project_engine(project)
    session = session_maker()

    collected = CommandLineRunner.collect(project, select_all=all_tests)

    for plugin_test in plugin_tests:
        try:
            plugin_name, test_name = plugin_test.split(":")
        except ValueError:
            plugin_name, test_name = plugin_test, None

        if test_name:
            collected[plugin_name].select_test(test_name)
        else:
            collected[plugin_name].select_all()

    exit_codes = asyncio.run(_run_plugin_tests(session, collected.values()))
    click.echo()
    _report_and_exit(exit_codes)


async def _run_plugin_tests(
    session: sessionmaker,
    runners: Iterable[ValidationsRunner],
) -> dict[str, dict[str, int]]:
    return {runner.plugin_name: await runner.run_all(session) for runner in runners}


def _report_and_exit(results: dict[str, dict[str, int]]):
    exit_code = 0
    failed_count = 0
    passed_count = 0

    for plugin_name, test_results in results.items():
        for test_name, test_exit_code in test_results.items():
            outcome = ValidationOutcome.from_exit_code(test_exit_code)
            if test_exit_code == 0:
                passed_count += 1
            else:
                exit_code = 1
                failed_count += 1
            styled_result = click.style(outcome, fg=outcome.color)
            plugin_test = click.style(f"{plugin_name}:{test_name}", bold=True)
            click.echo(f"{plugin_test.ljust(TEST_LINE_LENGTH, '.')} {styled_result}")

    status = "successfully" if failed_count == 0 else "with failures"
    message = (
        f"Testing completed {status}. "
        + f"{passed_count} test(s) successful. {failed_count} test(s) failed."
    )

    write_sep_line(message, "=", fg=("red" if exit_code else "green"))

    sys.exit(exit_code)



"""Project Settings Service."""

from __future__ import annotations

import json

import structlog
from dotenv import dotenv_values

from meltano.core.project import ProjectReadonly
from meltano.core.setting_definition import SettingDefinition
from meltano.core.settings_service import SettingsService, SettingValueStore
from meltano.core.utils import nest_object

from .config_service import ConfigService

logger = structlog.get_logger(__name__)

UI_CFG_SETTINGS = {
    "ui.server_name": "SERVER_NAME",
    "ui.secret_key": "SECRET_KEY",
    "ui.password_salt": "SECURITY_PASSWORD_SALT",
}


class ProjectSettingsService(SettingsService):
    """Project Settings Service."""

    config_override = {}
    supports_environments = False

    def __init__(self, *args, config_service: ConfigService = None, **kwargs):
        """Instantiate ProjectSettingsService instance.

        Args:
            args: Positional arguments to pass to the superclass.
            config_service: Project configuration service instance.
            kwargs: Keyword arguments to pass to the superclass.
        """
        super().__init__(*args, **kwargs)

        self.config_service = config_service or ConfigService(self.project)

        self.env_override = {
            # terminal environment variables already present from SettingService.env
            **self.project.env,  # static, project-level envs (e.g. MELTANO_ENVIRONMENT)
            **self.project.meltano.env,  # env vars stored in the base `meltano.yml` `env:` key
            **self.env_override,  # overrides
        }

        self.config_override = {  # noqa: WPS601
            **self.__class__.config_override,
            **self.config_override,
        }

        try:
            self.ensure_project_id()
        except ProjectReadonly:
            logger.debug(
                "Cannot update `project_id` in `meltano.yml`: project is read-only."
            )

    def ensure_project_id(self) -> None:
        """Ensure `project_id` is configured properly.

        Every `meltano.yml` file should contain the `project_id` key-value pair. It should be
        present in the top-level config, rather than in any environment-level configs.

        If it is not present, it will be restored from `analytics.json` if possible.
        """
        try:
            project_id = self.get("project_id")
        except OSError:
            project_id = None

        if project_id is None:
            try:
                with open(
                    self.project.meltano_dir() / "analytics.json"
                ) as analytics_json_file:
                    project_id = json.load(analytics_json_file)["project_id"]
            except (OSError, KeyError, json.JSONDecodeError) as err:
                logger.debug(
                    "Unable to restore 'project_id' from 'analytics.json'", err=err
                )
            else:
                self.set("project_id", project_id, store=SettingValueStore.MELTANO_YML)
                logger.debug("Restored 'project_id' from 'analytics.json'")

    @property
    def label(self) -> str:
        """Return label.

        Returns:
            Project label.
        """
        return "Meltano"

    @property
    def docs_url(self) -> str:
        """Return docs URL.

        Returns:
            URL for Meltano doc site.
        """
        return "https://docs.meltano.com/reference/settings"

    @property
    def db_namespace(self) -> str:
        """Return namespace for setting value records in system database.

        Returns:
            Namespace for setting value records in system database.
        """
        return "meltano"

    @property
    def setting_definitions(self) -> list[SettingDefinition]:
        """Return definitions of supported settings.

        Returns:
            A list of defined settings.
        """
        return self.config_service.settings

    @property
    def meltano_yml_config(self):
        """Return current configuration in `meltano.yml`.

        Returns:
            Current configuration in `meltano.yml`.
        """
        return self.config_service.current_config

    def update_meltano_yml_config(self, config):
        """Update configuration in `meltano.yml`.

        Args:
            config: Updated config.
        """
        self.config_service.update_config(config)

    def process_config(self, config) -> dict:
        """Process configuration dictionary for presentation in `meltano config meltano`.

        Args:
            config: Config to process.

        Returns:
            Processed configuration dictionary for presentation in `meltano config meltano`.
        """
        return nest_object(config)

    def get_with_metadata(self, name: str, *args, **kwargs):
        """Return setting value with metadata.

        Args:
            name: Name of setting to get.
            args: Positional arguments to pass to the superclass method.
            kwargs: Keyword arguments to pass to the superclass method.

        Returns:
            Setting value with metadata.
        """
        value, metadata = super().get_with_metadata(name, *args, **kwargs)
        source = metadata["source"]

        if source is SettingValueStore.DEFAULT:
            # Support legacy `ui.cfg` files generated by `meltano ui setup`
            ui_cfg_value = self.get_from_ui_cfg(name)
            if ui_cfg_value is not None:
                value = ui_cfg_value
                metadata["source"] = SettingValueStore.ENV
        return value, metadata

    def get_from_ui_cfg(self, name: str):
        """Return setting value from UI config.

        Args:
            name: Name of setting to get.

        Returns:
            Setting value from UI config.
        """
        try:
            key = UI_CFG_SETTINGS[name]
            config = dotenv_values(self.project.root_dir("ui.cfg"))
            value = config[key]

            # Since `ui.cfg` is technically a Python file, `'None'` means `None`
            if value == "None":
                value = None

            return value
        except KeyError:
            return None



"""Config management CLI."""
from __future__ import annotations

import asyncio
import json
import logging
import tempfile
from pathlib import Path

import click
import dotenv

from meltano.core.db import project_engine
from meltano.core.plugin import PluginType
from meltano.core.plugin.error import PluginNotFoundError
from meltano.core.plugin.settings_service import PluginSettingsService
from meltano.core.plugin_invoker import PluginInvoker
from meltano.core.plugin_test_service import PluginTestServiceFactory
from meltano.core.project import Project
from meltano.core.project_plugins_service import ProjectPluginsService
from meltano.core.project_settings_service import ProjectSettingsService
from meltano.core.settings_service import SettingValueStore
from meltano.core.settings_store import StoreNotSupportedError
from meltano.core.tracking import CliEvent, PluginsTrackingContext

from . import cli
from .params import pass_project
from .utils import CliError, InstrumentedGroup, PartialInstrumentedCmd

logger = logging.getLogger(__name__)


def get_label(metadata) -> str:
    """Get the label for an environment variable's source.

    Args:
        metadata: the metadata for the variable

    Returns:
        string describing the source of the variable's value
    """
    source = metadata["source"]
    try:
        return f"from the {metadata['env_var']} variable in {source.label}"
    except KeyError:
        return f"from {source.label}"


@cli.group(
    cls=InstrumentedGroup,
    invoke_without_command=True,
    short_help="Display Meltano or plugin configuration.",
)
@click.option(
    "--plugin-type", type=click.Choice(PluginType.cli_arguments()), default=None
)
@click.argument("plugin_name")
@click.option(
    "--format",
    "config_format",
    type=click.Choice(["json", "env"]),
    default="json",
)
@click.option("--extras", is_flag=True, help="View or list only plugin extras.")
@pass_project(migrate=True)
@click.pass_context
def config(  # noqa: WPS231
    ctx,
    project: Project,
    plugin_type: str,
    plugin_name: str,
    config_format: str,
    extras: bool,
):
    """
    Display Meltano or plugin configuration.

    \b\nRead more at https://docs.meltano.com/reference/command-line-interface#config
    """
    tracker = ctx.obj["tracker"]
    try:
        plugin_type = PluginType.from_cli_argument(plugin_type) if plugin_type else None
    except ValueError:
        tracker.track_command_event(CliEvent.aborted)
        raise

    if ctx.obj["is_default_environment"]:
        logger.info(
            f"The default environment ({project.active_environment.name}) will be ignored for `meltano config`. "
            + "To configure a specific Environment, please use option `--environment=<environment name>`."
        )
        project.deactivate_environment()

    plugins_service = ProjectPluginsService(project)

    try:
        plugin = plugins_service.find_plugin(
            plugin_name, plugin_type=plugin_type, configurable=True
        )
    except PluginNotFoundError:
        if plugin_name == "meltano":
            plugin = None
        else:
            tracker.track_command_event(CliEvent.aborted)
            raise

    if plugin:
        tracker.add_contexts(PluginsTrackingContext([(plugin, None)]))
    tracker.track_command_event(CliEvent.inflight)

    _, Session = project_engine(project)  # noqa: N806
    session = Session()
    try:
        if plugin:
            settings = PluginSettingsService(
                project,
                plugin,
                plugins_service=plugins_service,
            )
            invoker = PluginInvoker(project, plugin)
        else:
            settings = ProjectSettingsService(
                project, config_service=plugins_service.config_service
            )
            invoker = None

        ctx.obj["settings"] = settings
        ctx.obj["session"] = session
        ctx.obj["invoker"] = invoker

        if ctx.invoked_subcommand is None:
            if config_format == "json":
                process = extras is not True
                json_config = settings.as_dict(
                    extras=extras, process=process, session=session
                )
                click.echo(json.dumps(json_config, indent=2))
            elif config_format == "env":
                env = settings.as_env(extras=extras, session=session)

                with tempfile.NamedTemporaryFile() as temp_dotenv:
                    path = temp_dotenv.name
                    for key, value in env.items():
                        dotenv.set_key(path, key, value)

                    dotenv_content = Path(temp_dotenv.name).read_text()

                click.echo(dotenv_content)
    except Exception:
        tracker.track_command_event(CliEvent.failed)
        raise
    finally:
        session.close()


@config.command(
    cls=PartialInstrumentedCmd,
    name="list",
    short_help=(
        "List all settings for the specified plugin with their names, environment variables, and current values."
    ),
)
@click.option("--extras", is_flag=True)
@click.pass_context
def list_settings(ctx, extras: bool):
    """List all settings for the specified plugin with their names, environment variables, and current values."""
    settings = ctx.obj["settings"]
    session = ctx.obj["session"]
    tracker = ctx.obj["tracker"]

    printed_custom_heading = False
    printed_extra_heading = extras

    # If `--extras` is not specified (`False`), we still want to load both
    # regular and extra settings, since we show custom extras.
    load_extras = True if extras else None

    full_config = settings.config_with_metadata(session=session, extras=load_extras)
    for name, config_metadata in full_config.items():
        value = config_metadata["value"]
        source = config_metadata["source"]
        setting_def = config_metadata["setting"]

        if extras:
            if not setting_def.is_extra:
                continue

            if setting_def.is_custom and not printed_custom_heading:
                click.echo()
                click.echo("Custom:")
                printed_custom_heading = True
        elif setting_def.is_extra:
            if not setting_def.is_custom:
                continue

            if not printed_extra_heading:
                click.echo()
                click.echo("Custom extras, plugin-specific options handled by Meltano:")
                printed_extra_heading = True
        elif setting_def.is_custom and not printed_custom_heading:
            click.echo()
            click.echo("Custom, possibly unsupported by the plugin:")
            printed_custom_heading = True

        click.secho(name, fg="blue", nl=False)

        env_keys = [var.definition for var in settings.setting_env_vars(setting_def)]
        click.echo(f" [env: {', '.join(env_keys)}]", nl=False)

        if source is not SettingValueStore.DEFAULT:
            default_value = setting_def.value
            if default_value is not None:
                click.echo(f" (default: {default_value!r})", nl=False)

        if source is SettingValueStore.DEFAULT:
            label = "default"
        elif source is SettingValueStore.INHERITED:
            label = f"inherited from '{settings.plugin.parent.name}'"
        else:
            label = f"{get_label(config_metadata)}"

        current_value = click.style(f"{value!r}", fg="green")
        click.echo(f" current value: {current_value}", nl=False)

        unexpanded_value = config_metadata.get("unexpanded_value")
        if not unexpanded_value or unexpanded_value == value:
            click.echo(f" ({label})")
        else:
            click.echo(f" ({label}: {unexpanded_value!r})")

        if setting_def.description:
            click.echo("\t", nl=False)
            if setting_def.label:
                click.echo(f"{setting_def.label}: ", nl=False)
            click.echo(f"{setting_def.description}")

    docs_url = settings.docs_url
    if docs_url:
        click.echo()
        click.echo(
            f"To learn more about {settings.label} and its settings, visit {docs_url}"
        )
    tracker.track_command_event(CliEvent.completed)


@config.command(cls=PartialInstrumentedCmd)
@click.option(
    "--store",
    type=click.Choice(SettingValueStore.writables()),
    default=SettingValueStore.AUTO,
)
@click.pass_context
def reset(ctx, store):
    """Clear the configuration (back to defaults)."""
    store = SettingValueStore(store)

    settings = ctx.obj["settings"]
    session = ctx.obj["session"]
    tracker = ctx.obj["tracker"]

    try:
        metadata = settings.reset(store=store, session=session)
    except StoreNotSupportedError as err:
        tracker.track_command_event(CliEvent.aborted)
        raise CliError(
            f"{settings.label.capitalize()} settings in {store.label} could not be reset: {err}"
        ) from err

    store = metadata["store"]
    click.secho(
        f"{settings.label.capitalize()} settings in {store.label} were reset",
        fg="green",
    )
    tracker.track_command_event(CliEvent.completed)


@config.command(cls=PartialInstrumentedCmd, name="set")
@click.argument("setting_name", nargs=-1, required=True)
@click.argument("value")
@click.option(
    "--store",
    type=click.Choice(SettingValueStore.writables()),
    default=SettingValueStore.AUTO,
)
@click.pass_context
def set_(ctx, setting_name, value, store):
    """Set the configurations' setting `<name>` to `<value>`."""
    store = SettingValueStore(store)
    tracker = ctx.obj["tracker"]
    try:
        value = json.loads(value)
    except json.JSONDecodeError:
        pass

    settings = ctx.obj["settings"]
    session = ctx.obj["session"]

    path = list(setting_name)
    try:
        value, metadata = settings.set_with_metadata(
            path, value, store=store, session=session
        )
    except StoreNotSupportedError as err:
        tracker.track_command_event(CliEvent.aborted)
        raise CliError(
            f"{settings.label.capitalize()} setting '{path}' could not be set in {store.label}: {err}"
        ) from err

    name = metadata["name"]
    store = metadata["store"]
    click.secho(
        f"{settings.label.capitalize()} setting '{name}' was set in {store.label}: {value!r}",
        fg="green",
    )

    current_value, metadata = settings.get_with_metadata(name, session=session)
    source = metadata["source"]
    if source != store:
        message = f"Current value is still: {current_value!r} {get_label(metadata)}"
        click.secho(
            message,
            fg="yellow",
        )
    tracker.track_command_event(CliEvent.completed)


@config.command(cls=PartialInstrumentedCmd, name="test")
@click.pass_context
def test(ctx):
    """Test the configuration of a plugin."""
    invoker = ctx.obj["invoker"]
    tracker = ctx.obj["tracker"]
    if not invoker:
        tracker.track_command_event(CliEvent.aborted)
        raise CliError("Testing of the Meltano project configuration is not supported")

    session = ctx.obj["session"]

    async def _validate():  # noqa: WPS430
        plugin_test_service = PluginTestServiceFactory(invoker).get_test_service()
        async with plugin_test_service.plugin_invoker.prepared(session):
            return await plugin_test_service.validate()

    try:
        is_valid, detail = asyncio.run(_validate())
    except Exception:
        tracker.track_command_event(CliEvent.failed)
        raise

    if not is_valid:
        tracker.track_command_event(CliEvent.failed)
        raise CliError("\n".join(("Plugin configuration is invalid", detail)))

    click.secho("Plugin configuration is valid", fg="green")
    tracker.track_command_event(CliEvent.completed)


@config.command(cls=PartialInstrumentedCmd)
@click.argument("setting_name", nargs=-1, required=True)
@click.option(
    "--store",
    type=click.Choice(SettingValueStore.writables()),
    default=SettingValueStore.AUTO,
)
@click.pass_context
def unset(ctx, setting_name, store):
    """Unset the configurations' setting called `<name>`."""
    store = SettingValueStore(store)

    settings = ctx.obj["settings"]
    session = ctx.obj["session"]
    tracker = ctx.obj["tracker"]

    path = list(setting_name)
    try:
        metadata = settings.unset(path, store=store, session=session)
    except StoreNotSupportedError as err:
        tracker.track_command_event(CliEvent.aborted)
        raise CliError(
            f"{settings.label.capitalize()} setting '{path}' in {store.label} could not be unset: {err}"
        ) from err

    name = metadata["name"]
    store = metadata["store"]
    click.secho(
        f"{settings.label.capitalize()} setting '{name}' in {store.label} was unset",
        fg="green",
    )

    current_value, source = settings.get_with_source(name, session=session)
    if source is not SettingValueStore.DEFAULT:
        click.secho(
            f"Current value is now: {current_value!r} ({get_label(metadata)})",
            fg="yellow",
        )
    tracker.track_command_event(CliEvent.completed)



from __future__ import annotations

import cProfile
import logging
import pstats
import time

from flask import g as global_app_ctx
from sqlalchemy import event
from sqlalchemy.engine import Engine

logger = logging.getLogger("meltano.api.sql")
logger.setLevel(logging.DEBUG)


@event.listens_for(Engine, "before_cursor_execute")
def before_cursor_execute(conn, cursor, statement, parameters, context, executemany):
    conn.info.setdefault("query_start_time", []).append(time.time())
    logger.debug(f"Start Query: {statement}")


@event.listens_for(Engine, "after_cursor_execute")
def after_cursor_execute(conn, cursor, statement, parameters, context, executemany):
    total = time.time() - conn.info["query_start_time"].pop(-1)
    logger.debug("Query Complete!")
    logger.debug(f"Total Time: {total}")


def begin_request_profile():
    global_app_ctx.pr = cProfile.Profile()
    global_app_ctx.pr.enable()


def print_request_profile(res):
    global_app_ctx.pr.disable()
    global_app_ctx.pr.create_stats()

    stats = pstats.Stats(global_app_ctx.pr)
    stats.sort_stats("cumulative").print_stats(0.1)

    return res


def init(app):
    app.before_request(begin_request_profile)
    app.after_request(print_request_profile)



"""Schema creation CLI."""

from __future__ import annotations

import click

from meltano.core.db import ensure_schema_exists, project_engine

from . import cli
from .params import pass_project
from .utils import InstrumentedCmd, InstrumentedGroup


@cli.group(cls=InstrumentedGroup, short_help="Manage system DB schema.")
def schema():
    """Manage system DB schema."""


@schema.command(
    cls=InstrumentedCmd, short_help="Create system DB schema, if not exists."
)
@click.argument("schema")
@click.argument("roles", nargs=-1, required=True)
@pass_project()
def create(project, schema_name, roles):
    """Create system DB schema, if not exists."""
    engine, _ = project_engine(project)
    ensure_schema_exists(engine, schema_name, grant_roles=roles)



"""OAuth providers."""

from __future__ import annotations

from authlib.integrations.flask_client import OAuth as OAuthClient
from flask import Blueprint, Flask, render_template, url_for

OAuth = OAuthClient()


def facebook(app: Flask) -> None:
    """Register OAuth for Facebook.

    Args:
        app: The Flask app for which the OAuth blueprint will be registered.
    """
    OAuth.register(  # noqa: S106
        "facebook",
        access_token_url="https://graph.facebook.com/v5.0/oauth/access_token",
        client_kwargs={"scope": "ads_read ads_management manage_pages"},
        authorize_url="https://www.facebook.com/v5.0/dialog/oauth",
    )

    oauth_bp = Blueprint("OAuth-Facebook", __name__, url_prefix="/facebook")

    @oauth_bp.route("/", strict_slashes=False)
    def login():
        redirect_uri = url_for(".authorize", _external=True)
        return OAuth.facebook.authorize_redirect(
            redirect_uri, display="popup", auth_type="rerequest"
        )

    @oauth_bp.route("/authorize")
    def authorize():
        token = OAuth.facebook.authorize_access_token()
        return render_template("token.html", token=token["access_token"])

    app.register_blueprint(oauth_bp)


def google_adwords(app: Flask) -> None:
    """Register OAuth for Google AdWords.

    Args:
        app: The Flask app for which the OAuth blueprint will be registered.
    """
    OAuth.register(  # noqa: S106
        "google_adwords",
        authorize_url="https://accounts.google.com/o/oauth2/auth",
        client_kwargs={"scope": "https://www.googleapis.com/auth/adwords"},
        access_token_url="https://oauth2.googleapis.com/token",
    )

    oauth_bp = Blueprint("OAuth-GoogleAdwords", __name__, url_prefix="/google-adwords")

    @oauth_bp.route("/", strict_slashes=False)
    def login():
        redirect_uri = url_for(".authorize", _external=True)
        # `consent` prompt is required to have a refresh_token
        # `offline` access_type is required to have a refresh_token
        return OAuth.google_adwords.authorize_redirect(
            redirect_uri, prompt="consent", access_type="offline"
        )

    @oauth_bp.route("/authorize")
    def authorize():
        token = OAuth.google_adwords.authorize_access_token()
        return render_template("token.html", token=token["refresh_token"])

    app.register_blueprint(oauth_bp)



"""Add `last_heartbeat_at` column to `job` table.

Revision ID: d135f52a6f49
Revises: e4fbabc3fed6
Create Date: 2021-01-14 12:45:55.821947

"""
import sqlalchemy as sa
from alembic import op

from meltano.migrations.utils.dialect_typing import (
    datetime_for_dialect,
    get_dialect_name,
)

# revision identifiers, used by Alembic.
revision = "d135f52a6f49"
down_revision = "e4fbabc3fed6"
branch_labels = None
depends_on = None


def upgrade():
    dialect_name = get_dialect_name()
    datetime_type = datetime_for_dialect(dialect_name)

    op.add_column("job", sa.Column("last_heartbeat_at", datetime_type, nullable=True))


def downgrade():
    op.drop_column("user", "last_heartbeat_at")



"""Defines helpers for the core codebase."""

from __future__ import annotations

import asyncio
import functools
import hashlib
import logging
import math
import os
import re
import traceback
from collections import OrderedDict
from copy import deepcopy
from datetime import date, datetime, time
from pathlib import Path
from typing import Any, Callable, Iterable, TypeVar

import flatten_dict
from requests.auth import HTTPBasicAuth

logger = logging.getLogger(__name__)
T = TypeVar("T")

TRUTHY = ("true", "1", "yes", "on")
REGEX_EMAIL = r"(^[a-zA-Z0-9_.+-]+@[a-zA-Z0-9-]+\.[a-zA-Z0-9-.]+$)"


try:
    # asyncio.Task.all_tasks() is fully moved to asyncio.all_tasks() starting
    # with Python 3.9. Also applies to current_task.
    asyncio_all_tasks = asyncio.all_tasks
except AttributeError:
    asyncio_all_tasks = asyncio.Task.all_tasks


class NotFound(Exception):
    """Occurs when an element is not found."""

    def __init__(self, name, obj_type=None):
        """Create a new exception.

        Args:
            name: the name of the element that is not found
            obj_type: the type of element
        """
        if obj_type is None:
            super().__init__(f"{name} was not found.")
        else:
            super().__init__(f"{obj_type.__name__} '{name}' was not found.")


def click_run_async(func):
    """Small decorator to allow click invoked functions to leverage `asyncio.run` and be declared as async.

    Args:
        func: the function to run async

    Returns:
        A function which runs the given function async
    """

    @functools.wraps(func)
    def wrapper(*args, **kwargs):  # noqa: WPS430
        return asyncio.run(func(*args, **kwargs))

    return wrapper


# from https://github.com/jonathanj/compose/blob/master/compose.py
def compose(*fs: Callable[[Any], Any]):
    """Create a composition of unary functions.

    Args:
        fs: Unary functions to compose.

    Examples:
        ```python
        compose(f, g)(x) == f(g(x))
        ```

    Returns:
        The composition of the provided unary functions, which itself is a unary function.
    """
    return functools.reduce(lambda f, g: lambda x: f(g(x)), compact(fs), lambda x: x)


# from http://www.dolphmathews.com/2012/09/slugify-string-in-python.html
def slugify(s):
    """Normalize strings into something URL-friendly.

    Args:
        s: the string to slugify

    Returns:
        The string as a slug

    >>> slugify("[Some] _ Article's Title--")
    'some-articles-title'
    """
    # "[Some] _ Article's Title--" -> "[some] _ article's title--"
    s = s.lower()

    # "[some] _ article's_title--" -> "[some]___article's_title__"
    for c in " -./":
        s = s.replace(c, "_")

    # "[some]___article's_title__" -> "some___articles_title__"
    s = re.sub(r"\W", "", s)

    # "some___articles_title__" -> "some   articles title  "
    s = s.replace("_", " ")

    # "some   articles title  " -> "some articles title "
    s = re.sub(r"\s+", " ", s)

    # "some articles title " -> "some articles title"
    s = s.strip()

    # "some articles title" -> "some-articles-title"
    return s.replace(" ", "-")


def get_basic_auth(user, token):
    return HTTPBasicAuth(user, token)


def pop_all(keys, d: dict):
    return {k: d.pop(k) for k in keys}


def get_all(keys, d: dict, default=None):
    return {k: d.get(k, default) for k in keys}


# Taken from https://stackoverflow.com/a/20666342
def merge(src, dest):
    """Merge both given dictionaries together at depth, modifying `dest` in-place.

    Args:
        src: A dictionary to merge into `dest`.
        dest: The dictionary that will be updated with the keys and values from
            `src` at depth.

    Examples:
        >>> a = { 'first' : { 'all_rows' : { 'pass' : 'dog', 'number' : '1' } } }
        >>> b = { 'first' : { 'all_rows' : { 'fail' : 'cat', 'number' : '5' } } }
        >>> merge(b, a) == { 'first' : { 'all_rows' : { 'pass' : 'dog', 'fail' : 'cat', 'number' : '5' } } }
        True

    Returns:
        The `dest` dictionary with the keys and values from `src` merged in.
    """
    for key, value in src.items():
        if isinstance(value, dict):
            # get node or create one
            node = dest.setdefault(key, {})
            merge(value, node)
        else:
            dest[key] = value

    return dest


def nest(d: dict, path: str, value=None, maxsplit=-1, force=False):
    """Create a hierarchical dictionary path and return the leaf dict.

    Args:
        d: the dictionary to operate on
        path: the dot-delimited path to operate on
        value: the value to set at the given path
        maxsplit: maximum number of splits to split path by
        force: if true, write an empty dict

    Returns:
        The leaf element of the dict

    Examples:
        >>> d = dict()
        >>> test = nest(d, "foo.bar.test")
        >>> test
        {}
        >>> d
        {'foo': {'bar': {'test': {}}}}
        >>> test["a"] = 1
        >>> d
        {'foo': {'bar': {'test': {'a': 1}}}}
        >>> alist = nest(d, "foo.bar.list", value=[])
        >>> alist.append("works")
        >>> d
        {'foo': {'bar': {'test': {'a': 1}}, 'list': ["works"]}}
    """  # noqa: P102
    if value is None:
        value = {}

    if isinstance(path, str):
        path = path.split(".", maxsplit=maxsplit)

    *initial, tail = path

    # create the list of dicts
    cursor = d
    for key in initial:
        if key not in cursor or (not isinstance(cursor[key], dict) and force):
            cursor[key] = {}

        cursor = cursor[key]

    if tail not in cursor or (
        type(cursor[tail]) is not type(value) and force  # noqa: WPS516
    ):
        # We need to copy the value to make sure
        # the `value` parameter is not mutated.
        cursor[tail] = deepcopy(value)

    return cursor[tail]


def nest_object(flat_object):
    obj = {}
    for key, value in flat_object.items():
        nest(obj, key, value)
    return obj


def to_env_var(*xs):
    xs = [re.sub("[^A-Za-z0-9]", "_", x).upper() for x in xs if x]
    return "_".join(xs)


def flatten(d: dict, reducer: str | Callable = "tuple", **kwargs):
    """Flatten a dictionary with `dot` and `env_var` reducers.

    Wrapper arround `flatten_dict.flatten`.

    Args:
        d: the dict to flatten
        reducer: the reducer to flatten with
        **kwargs: additional kwargs to pass to flatten_dict.flatten

    Returns:
        the flattened dict
    """

    def dot_reducer(*xs):
        if xs[0] is None:
            return xs[1]
        return ".".join(xs)

    if reducer == "dot":
        reducer = dot_reducer
    if reducer == "env_var":
        reducer = to_env_var

    return flatten_dict.flatten(d, reducer, **kwargs)


def compact(xs: Iterable) -> Iterable:
    """Remove None values from an iterable.

    Args:
        xs: the iterable to operate on

    Returns:
        The iterable with Nones removed
    """
    return (x for x in xs if x is not None)


def file_has_data(file: Path | str):
    file = Path(file)  # ensure it is a Path object
    return file.exists() and file.stat().st_size > 0


def identity(x):
    return x


def noop(*_args, **_kwargs):
    pass


def map_dict(f: Callable, d: dict):
    yield from ((k, f(v)) for k, v in d.items())


def truthy(val: str) -> bool:
    return str(val).lower() in TRUTHY


def coerce_datetime(d: date | datetime) -> datetime | None:
    """Add a `time` component to `d` if it is missing.

    Args:
        d: the date or datetime to add the time to

    Returns:
        The resulting datetime
    """
    if d is None:
        return None

    if isinstance(d, datetime):
        return d

    return datetime.combine(d, time())


def iso8601_datetime(d: str) -> datetime | None:
    if d is None:
        return None

    isoformats = [
        "%Y-%m-%dT%H:%M:%SZ",  # noqa: WPS323
        "%Y-%m-%dT%H:%M:%S",  # noqa: WPS323
        "%Y-%m-%d %H:%M:%S",  # noqa: WPS323
        "%Y-%m-%d",  # noqa: WPS323
    ]

    for format_string in isoformats:
        try:
            return coerce_datetime(datetime.strptime(d, format_string))
        except ValueError:
            pass

    raise ValueError(f"{d} is not a valid UTC date.")


def find_named(xs: Iterable[dict], name: str, obj_type: type = None) -> dict:
    """Find an object by its 'name' key.

    Args:
        xs: Some iterable of objects against which that name should be matched.
        name: Used to match against the input objects.
        obj_type: Object type used for generating the exception message.

    Returns:
        The first item matched, if any. Otherwise raises an exception.

    Raises:
        NotFound: If an object with the given name was not found.
    """
    try:
        return next(x for x in xs if x["name"] == name)
    except StopIteration as stop:
        raise NotFound(name, obj_type) from stop


def makedirs(func):
    @functools.wraps(func)
    def decorate(*args, **kwargs):

        enabled = kwargs.get("make_dirs", True)

        path = func(*args, **kwargs)

        if not enabled:
            return path

        # if there is an extension, only create the base dir
        _, ext = os.path.splitext(path)
        if ext:
            directory = os.path.dirname(path)
        else:
            directory = path

        os.makedirs(directory, exist_ok=True)
        return path

    return decorate


def is_email_valid(value: str):
    return re.match(REGEX_EMAIL, value)


def pop_at_path(d, path, default=None):
    if isinstance(path, str):
        path = path.split(".")

    *initial, tail = path

    cursor = d
    cursors = []
    for key in initial:
        cursors.append((cursor, key))

        try:
            cursor = cursor[key]
        except KeyError:
            return default

    popped = cursor.pop(tail, default)

    for (cursor, key) in reversed(cursors):
        if len(cursor[key]) == 0:
            cursor.pop(key, None)

    return popped


def set_at_path(d, path, value):
    if isinstance(path, str):
        path = path.split(".")

    *initial, tail = path

    final = nest(d, initial, force=True) if initial else d
    final[tail] = value


class EnvironmentVariableNotSetError(Exception):
    """Occurs when a referenced environment variable is not set."""

    def __init__(self, env_var: str):
        """Initialize the error.

        Args:
            env_var: The unset environment variable name.
        """
        super().__init__(env_var)
        self.env_var = env_var

    def __str__(self) -> str:
        """Return the error as a string."""  # noqa: DAR201
        return f"{self.env_var} referenced but not set."


def expand_env_vars(raw_value, env: dict, raise_if_missing: bool = False):
    if isinstance(raw_value, dict):
        return {
            key: expand_env_vars(val, env, raise_if_missing)
            for key, val in raw_value.items()
        }
    elif not isinstance(raw_value, str):
        return raw_value

    # find viable substitutions
    var_matcher = re.compile(
        r"""
        \$  # starts with a '$'
        (?:
            {(\w+)} # ${VAR}
            |
            ([A-Z][A-Z0-9_]*) # $VAR
        )
        """,
        re.VERBOSE,
    )

    def subst(match) -> str:
        try:
            # the variable can be in either group
            var = next(var for var in match.groups() if var)
            val = str(env[var])

            if not val:
                logger.debug(f"Variable '${var}' is empty.")
                if raise_if_missing:
                    raise EnvironmentVariableNotSetError(var)
            return val
        except KeyError as e:
            if raise_if_missing:
                raise EnvironmentVariableNotSetError(e.args[0])
            logger.debug(f"Variable '${var}' is missing from the environment.")
            return None

    fullmatch = re.fullmatch(var_matcher, raw_value)
    if fullmatch:
        # If the entire value is an env var reference, return None if it isn't set
        return subst(fullmatch)

    return re.sub(var_matcher, subst, raw_value)


def uniques_in(original):
    return list(OrderedDict.fromkeys(original))


# https://gist.github.com/cbwar/d2dfbc19b140bd599daccbe0fe925597#gistcomment-2845059
def human_size(num, suffix="B"):
    """Return human-readable file size.

    Args:
        num: the number to convert
        suffix: the suffix to append to the resulting file size

    Returns:
        File size in human-readable format
    """
    magnitude = int(math.floor(math.log(num, 1024)))
    val = num / math.pow(1024, magnitude)

    if magnitude == 0:
        return f"{val:.0f} bytes"
    if magnitude > 7:
        return f"{val:.1f}Yi{suffix}"

    prefix = ["", "Ki", "Mi", "Gi", "Ti", "Pi", "Ei", "Zi"][magnitude]
    return f"{val:3.1f}{prefix}{suffix}"


def hash_sha256(value: str | bytes) -> str:
    """Get the sha256 hash of a string.

    Args:
        value: the string value to hash.

    Returns:
        The hashed value of the given string.

    Raises:
        ValueError: If we are blindly passed a value that is None.
    """
    if value is None:
        raise ValueError("Cannot hash None.")
    if isinstance(value, str):
        value = value.encode()
    return hashlib.sha256(value).hexdigest()


def format_exception(exception: BaseException) -> str:
    """Get the exception with its traceback in the standard format it would have been printed with.

    Args:
        exception: The exception value to be turned into a string.

    Returns:
        A string that shows the exception object as it would have been printed had it been raised
        and not caught.
    """
    return "".join(
        traceback.format_exception(type(exception), exception, exception.__traceback__)
    )


def safe_hasattr(obj: Any, name: str) -> bool:
    """Safely checks if an object has a given attribute.

    This is a hacky workaround for the fact that `hasattr` is not allowed by WPS.

    Args:
        obj: The object to check.
        name: The name of the attribute to check.

    Returns:
        True if the object has the attribute, False otherwise.
    """
    try:
        getattr(obj, name)
    except AttributeError:
        return False
    return True



"""Defines helpers related to the system database."""

from __future__ import annotations

import logging
import time

from sqlalchemy import create_engine
from sqlalchemy.engine import Connection, Engine
from sqlalchemy.exc import OperationalError
from sqlalchemy.orm import sessionmaker
from sqlalchemy.sql import text

from meltano.core.project import Project

from .project_settings_service import ProjectSettingsService

# Keep a Project → Engine mapping to serve
# the same engine for the same Project
_engines = {}


def project_engine(
    project: Project,
    default: bool = False,
) -> tuple[Engine, sessionmaker]:
    """Create and register a SQLAlchemy engine for a Meltano project instance.

    Args:
        project: The Meltano project that the engine will be connected to.
        default: Whether the engine created should be stored as the default
            engine for this project.

    Returns:
        The engine, and a session maker bound to the engine.
    """
    existing_engine = _engines.get(project)
    if existing_engine:
        return existing_engine

    settings = ProjectSettingsService(project)

    engine_uri = settings.get("database_uri")
    logging.debug(f"Creating engine {project}@{engine_uri}")
    engine = create_engine(engine_uri, pool_pre_ping=True)

    # Connect to the database to ensure it is available.
    connect(
        engine,
        max_retries=settings.get("database_max_retries"),
        retry_timeout=settings.get("database_retry_timeout"),
    )

    init_hook(engine)

    engine_session = (engine, sessionmaker(bind=engine))

    if default:
        # register the default engine
        _engines[project] = engine_session

    return engine_session


def connect(
    engine: Engine,
    max_retries: int,
    retry_timeout: float,
) -> Connection:
    """Connect to the database.

    Args:
        engine: The DB engine with which the check will be performed.
        max_retries: The maximum number of retries that will be attempted.
        retry_timeout: The number of seconds to wait between retries.

    Raises:
        OperationalError: Error during DB connection - max retries exceeded.

    Returns:
        A connection to the database.
    """
    attempt = 0
    while True:
        try:
            return engine.connect()
        except OperationalError:
            if attempt >= max_retries:
                logging.error(
                    f"Could not connect to the database after {attempt} "
                    "attempts. Max retries exceeded."
                )
                raise
            attempt += 1
            logging.info(
                f"DB connection failed. Will retry after {retry_timeout}s. "
                f"Attempt {attempt}/{max_retries}"
            )
            time.sleep(retry_timeout)


init_hooks = {
    "sqlite": lambda x: x.execute("PRAGMA journal_mode=WAL"),
}


def init_hook(engine: Engine) -> None:
    """Run the initialization hook for the provided DB engine.

    The initialization hooks are taken from the `meltano.core.db.init_hooks`
    dictionary, which maps the dialect name of the engine to a unary function
    which will be called with the provided DB engine.

    Args:
        engine: The engine for which the init hook will be run.

    Raises:
        Exception: The init hook raised an exception.
    """
    try:
        hook = init_hooks[engine.dialect.name]
    except KeyError:
        return

    try:
        hook(engine)
    except Exception as ex:
        raise Exception(f"Failed to initialize database: {ex!s}") from ex


def ensure_schema_exists(
    engine: Engine,
    schema_name: str,
    grant_roles: tuple[str] = (),
) -> None:
    """Ensure the specified `schema_name` exists in the database.

    Args:
        engine: The DB engine to be used.
        schema_name: The name of the schema.
        grant_roles: Roles to grant to the specified schema.
    """
    schema_identifier = schema_name
    group_identifiers = ",".join(grant_roles)

    create_schema = text(f"CREATE SCHEMA IF NOT EXISTS {schema_identifier}")
    grant_select_schema = text(
        f"ALTER DEFAULT PRIVILEGES IN SCHEMA {schema_identifier} GRANT SELECT ON TABLES TO {group_identifiers}"
    )
    grant_usage_schema = text(
        f"GRANT USAGE ON SCHEMA {schema_identifier} TO {group_identifiers}"
    )

    with engine.connect() as conn, conn.begin():
        conn.execute(create_schema)
        if grant_roles:
            conn.execute(grant_select_schema)
            conn.execute(grant_usage_schema)

    logging.info(f"Schema {schema_name} has been created successfully.")
    for role in grant_roles:
        logging.info(f"Usage has been granted for role: {role}.")



from __future__ import annotations

import logging

from blinker import ANY
from flask import render_template, url_for

from meltano.api.mail import MailService, mail
from meltano.api.models.subscription import Subscription, SubscriptionEventType
from meltano.api.signals import PipelineSignals
from meltano.core.plugin import PluginType
from meltano.core.project import Project
from meltano.core.project_plugins_service import ProjectPluginsService

SUCCESS = True
FAILURE = False

logger = logging.getLogger(__name__)


class NotificationEvents:
    def __init__(
        self,
        project: Project,
        mail_service: MailService = None,
        plugins_service: ProjectPluginsService = None,
    ):
        self.project = project

        self.mail_service = mail_service or MailService(project)
        self.plugins_service = plugins_service or ProjectPluginsService(project)

    def init_app(self, app):
        # wire the signal handlers
        PipelineSignals.completed.connect(self.handle_pipeline_completed, ANY)

    def pipeline_data_source(self, schedule) -> str:
        """Return the Data Source name for a Pipeline."""
        return self.plugins_service.find_plugin(
            schedule.extractor, plugin_type=PluginType.EXTRACTORS
        ).label

    def pipeline_urls(self, schedule) -> str:
        """Return external URLs to different point of interests for a Pipeline."""
        plugin = self.plugins_service.find_plugin(
            schedule.extractor, plugin_type=PluginType.EXTRACTORS
        )
        return {
            "log": url_for(
                "root.default", path=f"data/schedule/{schedule.name}", _external=True
            ),
            "config": url_for(
                "root.default",
                path=f"data/extract/{schedule.extractor}",
                _external=True,
            ),
            "docs": plugin.docs,
        }

    def handle_pipeline_completed(self, schedule, success: bool = None):  # noqa: WPS210
        """Handle the `Manual Run` pipeline email notification."""
        if success is None:
            raise ValueError("'success' must be set.")

        state_id = schedule.name
        data_source = self.pipeline_data_source(schedule)

        subject, template = {
            SUCCESS: (
                f"Your {data_source} data is ready!",
                "email/pipeline_manual_run/success.html",
            ),
            FAILURE: (
                f"Your {data_source} extraction has failed!",
                "email/pipeline_manual_run/failure.html",
            ),
        }[success]
        html = render_template(
            template, data_source=data_source, urls=self.pipeline_urls(schedule)
        )

        subscriptions = Subscription.query.filter_by(
            event_type=SubscriptionEventType.PIPELINE_MANUAL_RUN,
            source_type="pipeline",
            source_id=state_id,
        ).all()
        logger.debug(
            f"Found {len(subscriptions)} subscriptions for source '{state_id}'."
        )

        messages = set()
        sent_recipients = set()
        for subscription in subscriptions:
            # ensure it is not a duplicate mail
            if subscription.recipient in sent_recipients:
                logger.debug(
                    f"Skipping duplicate notification for recipient '{subscription.recipient}'."
                )
                continue

            msg = self.mail_service.create_message(
                subscription, subject=subject, html=html
            )
            sent_recipients.add(subscription.recipient)
            messages.add(msg)

        with mail.connect() as conn:
            for msg in messages:
                conn.send(msg)
                logger.debug(f"Notification sent to '{msg.recipients}'.")



"""Various small utilities for working with futures."""

from __future__ import annotations

import asyncio
import logging
from asyncio import Task

from meltano.core.runner import RunnerError
from meltano.core.utils import human_size


def all_done(tasks: list[Task], done: set[Task]) -> bool:
    """Iterate through a task list checking if ALL tasks are in the done set."""
    for idx in tasks:
        if idx not in done:
            return False
    return True


def first_failed_future(exception_future: Task, done: set[Task]) -> Task | None:
    """Check if a future is in a set of completed futures and return the first failed (if any).

    Args:
        exception_future: The future you want to check and who's futures should be returned if an exception was raised.
        done: The set of completed futures you want to search.

    Returns:
        The first future that failed.
    """
    if exception_future in done:
        futures_done, _ = exception_future.result()
        futures_failed = [
            future for future in futures_done if future.exception() is not None
        ]

        if futures_failed:
            return futures_failed.pop()


def handle_producer_line_length_limit_error(
    exception: Exception, line_length_limit: int, stream_buffer_size: int
):
    """
    Handle and wrap asyncio.LimitOverrunError's from producers, emitting an useful log line along the way.

    TODO: reuse from runner/singer.py
    StreamReader.readline can raise a ValueError wrapping a LimitOverrunError:
    https://github.com/python/cpython/blob/v3.8.7/Lib/asyncio/streams.py#L549
    """
    if not isinstance(exception, ValueError):
        return

    exception = exception.__context__  # noqa: WPS609
    if not isinstance(exception, asyncio.LimitOverrunError):
        return

    logging.error(
        f"The extractor generated a message exceeding the message size limit of {human_size(line_length_limit)} (half the buffer size of {human_size(stream_buffer_size)})."
    )
    logging.error(
        "To let this message be processed, increase the 'elt.buffer_size' setting to at least double the size of the largest expected message, and try again."
    )
    logging.error(
        "To learn more, visit https://docs.meltano.com/reference/settings#eltbuffer_size"
    )
    raise RunnerError("Output line length limit exceeded") from exception



"""Validates yaml files against their schemas."""

from __future__ import annotations

import json
import os
import pathlib

import yaml
from jsonschema import validate

proj_path = pathlib.Path(__file__).resolve().parents[4]

with open(os.path.join(proj_path, "schema/discovery.schema.json")) as schema_file:
    schema_content = json.load(schema_file)

with open(
    os.path.join(proj_path, "src/meltano/core/bundle/discovery.yml")
) as yaml_file:
    yaml_content = yaml.safe_load(yaml_file)
    validate(instance=yaml_content, schema=schema_content)



"""Snowplow contexts for Meltano telemetry."""

from __future__ import annotations

from .cli import CliContext, CliEvent
from .environment import environment_context
from .exception import ExceptionContext
from .plugins import PluginsTrackingContext
from .project import ProjectContext



"""Defines JobFinder (will be renamed to StateFinder)."""

from __future__ import annotations

from datetime import datetime, timedelta

from .job import HEARTBEAT_VALID_MINUTES, HEARTBEATLESS_JOB_VALID_HOURS, Job, State


class JobFinder:
    """Query builder for the `Job` model for a certain `elt_uri`."""

    def __init__(self, state_id: str):
        """Initialize the JobFinder.

        Args:
            state_id: the state_id to build queries for.
        """
        self.state_id = state_id

    def latest(self, session):
        """Get the latest state for this instance's state ID.

        Args:
            session: the session to use in querying the db

        Returns:
            The latest state for this instance's state ID
        """
        return (
            session.query(Job)
            .filter(Job.job_name == self.state_id)
            .order_by(Job.started_at.desc())
            .first()
        )

    def successful(self, session):
        """Get all successful jobs for this instance's state ID.

        Args:
            session: the session to use in querying the db

        Returns:
            All successful jobs for this instance's state ID
        """
        return session.query(Job).filter(
            (Job.job_name == self.state_id)  # noqa: WPS465
            & (Job.state == State.SUCCESS)  # noqa: WPS465
            & Job.ended_at.isnot(None)
        )

    def running(self, session):
        """Find states in the running state.

        Args:
            session: the session to use in querying the db

        Returns:
            All runnings states for state_id.
        """
        return session.query(Job).filter(
            (Job.job_name == self.state_id)  # noqa: WPS465
            & (Job.state == State.RUNNING)
        )

    def latest_success(self, session):
        """Get the latest successful state for this instance's state ID.

        Args:
            session: the session to use in querying the db

        Returns:
            The latest successful state for this instance's state ID
        """
        return self.successful(session).order_by(Job.ended_at.desc()).first()

    def latest_running(self, session):
        """Find the most recent state in the running state, if any.

        Args:
            session: the session to use in querying the db

        Returns:
            The latest running state for this instance's state ID
        """
        return self.running(session).order_by(Job.started_at.desc()).first()

    def with_payload(self, session, flags=0, since=None, state=None):
        """Get all states for this instance's state ID matching the given args.

        Args:
            session: the session to use in querying the db
            flags: only return states with these flags
            since: only return states which ended after this time
            state: only include states with state matching this state

        Returns:
            All states matching these args.
        """
        query = (
            session.query(Job)
            .filter(
                (Job.job_name == self.state_id)  # noqa: WPS465
                & (Job.payload_flags != 0)  # noqa: WPS465
                & (Job.payload_flags.op("&")(flags) == flags)  # noqa: WPS465
                & Job.ended_at.isnot(None)
            )
            .order_by(Job.ended_at.asc())
        )

        if since:
            query = query.filter(Job.ended_at > since)
        if state:
            query = query.filter(Job.state == state)
        return query

    def latest_with_payload(self, session, **kwargs):
        """Return the latest state matching the given kwargs.

        Args:
            session: the session to use to query the db
            kwargs: keyword args to pass to with_payload

        Returns:
            The most recent state returned by with_payload(kwargs)
        """
        return (
            self.with_payload(session, **kwargs)
            .order_by(None)  # Reset ascending order
            .order_by(Job.ended_at.desc())
            .first()
        )

    @classmethod
    def all_stale(cls, session):
        """Return all stale states.

        Args:
            session: the session to use to query the db

        Returns:
            All stale states with any state ID
        """
        now = datetime.utcnow()
        last_valid_heartbeat_at = now - timedelta(minutes=HEARTBEAT_VALID_MINUTES)
        last_valid_started_at = now - timedelta(hours=HEARTBEATLESS_JOB_VALID_HOURS)

        return session.query(Job).filter(
            (Job.state == State.RUNNING)  # noqa: WPS465
            & (
                (
                    Job.last_heartbeat_at.isnot(None)  # noqa: WPS465
                    & (Job.last_heartbeat_at < last_valid_heartbeat_at)
                )
                | (
                    Job.last_heartbeat_at.is_(None)  # noqa: WPS465
                    & (Job.started_at < last_valid_started_at)
                )
            )
        )

    def stale(self, session):
        """Return stale states with the instance's state ID.

        Args:
            session: the session to use in querying the db

        Returns:
            All stale states with instance's state ID
        """
        return self.all_stale(session).filter(Job.job_name == self.state_id)

    def get_all(self, session: object, since=None):
        """Return all state with the instance's state ID.

        Args:
            session: the session to use in querying the db
            since: only return state which ended after this datetime

        Returns:
            All state with instance's state ID which ended after 'since'
        """
        query = (
            session.query(Job)
            .filter(Job.job_name == self.state_id)
            .order_by(Job.ended_at.asc())
        )

        if since:
            query = query.filter(Job.ended_at > since)
        return query



"""State management in CLI."""
from __future__ import annotations

import json
import re
from datetime import datetime as dt
from functools import partial, reduce, wraps
from operator import xor

import click
import structlog

from meltano.cli.params import pass_project
from meltano.core.block.parser import BlockParser
from meltano.core.db import project_engine
from meltano.core.job import Payload
from meltano.core.project import Project
from meltano.core.state_service import InvalidJobStateError, StateService

from . import cli
from .utils import InstrumentedCmd, InstrumentedGroup

STATE_SERVICE_KEY = "state_service"

logger = structlog.getLogger(__name__)


class MutuallyExclusiveOptionsError(Exception):
    """Occurs when mutually exclusive options are provided incorrectly."""

    def __init__(self, *options: str) -> None:
        """Instantiate the error.

        Args:
            options: the mutually exclusive options that were incorrectly provided.
        """
        super().__init__(*options)
        self.options = options

    def __str__(self) -> str:
        """Represent the error as a string."""
        return f"Must provide exactly one of: {','.join(self.options)}"


def _prompt_for_confirmation(prompt):
    """Wrap destructive CLI commands which should prompt the user for confirmation."""

    def wrapper(func):
        fun = click.option(
            "--force", is_flag=True, help="Don't prompt for confirmation."
        )(func)

        @wraps(func)
        def _wrapper(force=False, *args, **kwargs):
            if force or click.confirm(prompt):
                return fun(*args, **kwargs, force=force)
            else:
                click.secho("Aborting.", fg="red")

        return _wrapper

    return wrapper


prompt_for_confirmation = partial(
    _prompt_for_confirmation, prompt="This is a destructive command. Continue?"
)


def state_service_from_state_id(project: Project, state_id: str) -> StateService | None:
    """Instantiate by parsing a state_id."""
    state_id_re = re.compile(r"^(?P<env>.+)\:(?P<tap>.+)-to-(?P<target>.+)$")
    match = state_id_re.match(state_id)
    if match:
        # If the state_id matches convention (i.e., job has been run via "meltano run"),
        # try parsing into BlockSet.
        # This way, we get BlockSet validation and raise an error if no
        # plugin in the BlockSet has "state" capability
        try:
            if not project.active_environment:
                logger.warn(
                    f"Running state operation for environment '{match.group('env')}' outside of an environment"
                )
            elif project.active_environment.name != match.group("env"):
                logger.warn(
                    f"Environment '{match.group('env')}' used in state operation does not match current environment '{project.active_environment.name}'."
                )
            project.activate_environment(match.group("env"))
            blocks = [match.group("tap"), match.group("target")]
            parser = BlockParser(logger, project, blocks)
            return next(parser.find_blocks()).state_service
        except Exception:
            logger.warn("No plugins found for provided state_id.")
    # If provided state_id does not match convention (i.e., run via "meltano elt"),
    # use the standalone StateService in the CLI context.
    return None


@cli.group(cls=InstrumentedGroup, name="state", short_help="Manage Singer state.")
@click.pass_context
@pass_project(migrate=True)
def meltano_state(project: Project, ctx: click.Context):
    """
    Manage state.

    \b\nRead more at https://docs.meltano.com/reference/command-line-interface#state
    """
    _, sessionmaker = project_engine(project)
    session = sessionmaker()
    ctx.obj[STATE_SERVICE_KEY] = StateService(session)  # noqa: WPS204


@meltano_state.command(cls=InstrumentedCmd, name="list")
@click.option("--pattern", type=str, help="Filter state IDs by pattern.")
@click.pass_context
@pass_project()
def list_state(
    project: Project, ctx: click.Context, pattern: str | None
):  # noqa: WPS125
    """List all state_ids for this project.

    Optionally pass a glob-style pattern to filter state_ids by.
    """
    state_service: StateService = ctx.obj[STATE_SERVICE_KEY]
    states = state_service.list_state(pattern)
    if states:
        for state_id, state in states.items():
            if state:
                try:
                    state_service.validate_state(json.dumps(state))
                except (InvalidJobStateError, json.decoder.JSONDecodeError):
                    click.secho(state_id, fg="red")
                else:
                    click.secho(state_id, fg="green")
            else:
                click.secho(state_id, fg="yellow")
    else:
        logger.info("No state IDs found.")


@meltano_state.command(cls=InstrumentedCmd, name="copy")
@prompt_for_confirmation(
    prompt="This will overwrite state for the destination. Continue?"
)
@click.argument("src-state-id", type=str)
@click.argument("dst-state-id", type=str)
@pass_project(migrate=True)
@click.pass_context
def copy_state(
    ctx: click.Context,
    project: Project,
    src_state_id: str,
    dst_state_id: str,
    force: bool,
):
    """Copy state to another job id."""
    # Retrieve state for copying
    state_service: StateService = (
        state_service_from_state_id(project, src_state_id) or ctx.obj[STATE_SERVICE_KEY]
    )

    state_service.copy_state(src_state_id, dst_state_id)

    logger.info(
        f"State for {dst_state_id} was successfully copied from {src_state_id} at {dt.utcnow():%Y-%m-%d %H:%M:%S}."  # noqa: WPS323
    )


@meltano_state.command(cls=InstrumentedCmd, name="move")
@prompt_for_confirmation(
    prompt="This will clear the source state and overwrite destination state. Continue?"
)
@click.argument("src-state-id", type=str)
@click.argument("dst-state-id", type=str)
@pass_project(migrate=True)
@click.pass_context
def move_state(
    ctx: click.Context,
    project: Project,
    src_state_id: str,
    dst_state_id: str,
    force: bool,
):
    """Move state to another job id, clearing the original."""
    # Retrieve state for moveing
    state_service: StateService = (
        state_service_from_state_id(project, dst_state_id) or ctx.obj[STATE_SERVICE_KEY]
    )

    state_service.move_state(src_state_id, dst_state_id)

    logger.info(
        f"State for {src_state_id} was successfully moved to {dst_state_id} at {dt.utcnow():%Y-%m-%d %H:%M:%S}."  # noqa: WPS323
    )


@meltano_state.command(cls=InstrumentedCmd, name="merge")
@click.option(
    "--from-state-id",
    type=str,
    help="Merge state from an existing state ID.",
)
@click.option(
    "--input-file",
    type=click.Path(exists=True),
    help="Merge state from a JSON file containing Singer state.",
)
@click.argument("state-id", type=str)
@click.argument("state", type=str, required=False)
@pass_project(migrate=True)
@click.pass_context
def merge_state(
    ctx: click.Context,
    project: Project,
    state_id: str,
    state: str | None,
    input_file: click.Path | None,
    from_state_id: str | None,
):
    """Add bookmarks to existing state."""
    state_service: StateService = (
        state_service_from_state_id(project, state_id) or ctx.obj[STATE_SERVICE_KEY]
    )
    mutually_exclusive_options = {
        "--input-file": input_file,
        "STATE": state,
        "--from-state-id": from_state_id,
    }
    if not reduce(xor, (bool(x) for x in mutually_exclusive_options.values())):
        raise MutuallyExclusiveOptionsError(*mutually_exclusive_options)
    elif input_file:
        with open(input_file) as state_f:
            state_service.add_state(
                state_id, state_f.read(), payload_flags=Payload.INCOMPLETE_STATE
            )
    elif state:
        state_service.add_state(state_id, state, payload_flags=Payload.INCOMPLETE_STATE)
    elif from_state_id:
        state_service.merge_state(from_state_id, state_id)
    logger.info(
        f"State for {state_id} was successfully merged at {dt.utcnow():%Y-%m-%d %H:%M:%S}."  # noqa: WPS323
    )


@meltano_state.command(cls=InstrumentedCmd, name="set")
@prompt_for_confirmation(
    prompt="This will overwrite the state's current value. Continue?"
)
@click.option(
    "--input-file",
    type=click.Path(exists=True),
    help="Set state from json file containing Singer state.",
)
@click.argument("state-id")
@click.argument("state", type=str, required=False)
@pass_project(migrate=True)
@click.pass_context
def set_state(
    ctx: click.Context,
    project: Project,
    state_id: str,
    state: str | None,
    input_file: click.Path | None,
    force: bool,
):
    """Set state."""
    state_service: StateService = (
        state_service_from_state_id(project, state_id) or ctx.obj[STATE_SERVICE_KEY]
    )
    mutually_exclusive_options = {
        "--input-file": input_file,
        "STATE": state,
    }
    if not reduce(xor, (bool(x) for x in mutually_exclusive_options.values())):
        raise MutuallyExclusiveOptionsError(*mutually_exclusive_options)
    elif input_file:
        with open(input_file) as state_f:
            state_service.set_state(state_id, state_f.read())
    elif state:
        state_service.set_state(state_id, state)
    logger.info(
        f"State for {state_id} was successfully set at {dt.utcnow():%Y-%m-%d %H:%M:%S}."  # noqa: WPS323
    )


@meltano_state.command(cls=InstrumentedCmd, name="get")  # noqa: WPS46
@click.argument("state-id")
@pass_project(migrate=True)
@click.pass_context
def get_state(ctx: click.Context, project: Project, state_id: str):  # noqa: WPS463
    """Get state."""
    state_service: StateService = (
        state_service_from_state_id(project, state_id) or ctx.obj[STATE_SERVICE_KEY]
    )
    retrieved_state = state_service.get_state(state_id)
    click.echo(json.dumps(retrieved_state))


@meltano_state.command(cls=InstrumentedCmd, name="clear")
@prompt_for_confirmation(prompt="This will clear state for the job. Continue?")
@click.argument("state-id")
@pass_project(migrate=True)
@click.pass_context
def clear_state(ctx: click.Context, project: Project, state_id: str, force: bool):
    """Clear state."""
    state_service: StateService = (
        state_service_from_state_id(project, state_id) or ctx.obj[STATE_SERVICE_KEY]
    )
    state_service.clear_state(state_id)



"""The actual `IOBlock` interface is one of the lower level blocks for use with various `BlockSet` implementations."""

from __future__ import annotations

from abc import ABCMeta, abstractmethod
from asyncio import StreamWriter, Task

from meltano.core.logging.utils import SubprocessOutputWriter


class IOBlock(metaclass=ABCMeta):
    """The IOBlock interface is a basic block that Consumes, Produces, or Consume and Produces (Transforms) output.

    Underlying implementation could be subprocesses (ala Singer Plugins), or stream transformers, basically,
    any class that satisfies the IOBlock interface.
    """

    @property
    @abstractmethod
    def stdin(self) -> StreamWriter | None:
        """If a block requires input, return the StreamWriter that should be used for writes.

        Returns:
            StreamWriter
        Raises:
            NotImplementedError
        """
        raise NotImplementedError

    @property
    @abstractmethod
    def consumer(self) -> bool:
        """Consumer indicates whether or not this block is a consumer and requires input."""
        raise NotImplementedError

    @property
    @abstractmethod
    def producer(self) -> bool:
        """Indicate whether or not this block is a producer of output."""
        raise NotImplementedError

    @property
    @abstractmethod
    def string_id(self) -> str:
        """Return a string identifier for this block."""
        raise NotImplementedError

    @property
    @abstractmethod
    def has_state(self) -> bool:
        """Indicate whether this block has persistent state between runs."""
        raise NotImplementedError

    @abstractmethod
    def stdout_link(self, dst: SubprocessOutputWriter) -> None:
        """Use stdout_link to instruct block to link/write stdout content to dst.

        Args:
            dst: SubprocessOutputWriter

        Raises:
            NotImplementedError
        """
        raise NotImplementedError

    @abstractmethod
    def stderr_link(self, dst: SubprocessOutputWriter) -> None:
        """Use stderr_link to instruct block to link/write stderr content to dst.

        Args:
            dst: SubprocessOutputWriter

        Raises:
            NotImplementedError
        """
        raise NotImplementedError

    @abstractmethod
    async def start(self) -> None:
        """Start the block.

        Whatever that might entail (spwaning a process, spinning up a async task that will handle transforms, etc)

        Raises:
            NotImplementedError
        """
        raise NotImplementedError

    @abstractmethod
    async def stop(self, kill: bool = True) -> None:
        """Stop a block.

        Args:
            kill: whether or not to send a SIGKILL. If false, a SIGTERM is sent.

        Raises:
            NotImplementedError
        """
        raise NotImplementedError

    @abstractmethod
    def proxy_stdout(self) -> Task:
        """Start proxying stdout to the linked stdout destinations.

        Returns: Future of the proxy task.
        """
        raise NotImplementedError

    @abstractmethod
    def proxy_stderr(self) -> Task:
        """Start proxying stderr to the linked stderr destinations.

        Returns: Future of the proxy task.
        """
        raise NotImplementedError

    @abstractmethod
    def proxy_io(self) -> tuple[Task, Task]:
        """Start proxying stdout AND stderr to the linked destinations.

        Returns:
            proxy_stdout Task and proxy_stderr Task
        """
        stdout = self.proxy_stdout()
        stderr = self.proxy_stderr()
        return stdout, stderr

    @abstractmethod
    async def pre(self, context: object) -> None:
        """Execute pre-start tasks.

        Args:
            context: invocation context to use for this execution.
        """

    @abstractmethod
    async def post(self) -> None:
        """Execute post-stop tasks."""

    @abstractmethod
    async def close_stdin(self) -> None:
        """Close the underlying stdin if the block is a consumer."""



"""Environment context for the Snowplow tracker."""

from __future__ import annotations

import os
import platform
import uuid
from collections import defaultdict
from datetime import datetime
from pathlib import Path
from typing import Any

import psutil
from cached_property import cached_property
from snowplow_tracker import SelfDescribingJson
from structlog.stdlib import get_logger

import meltano
from meltano.core.tracking.schemas import EnvironmentContextSchema
from meltano.core.utils import hash_sha256, safe_hasattr

logger = get_logger(__name__)

# This file is only ever created in CI when building a release
release_marker_path = Path(__file__).parent / ".release_marker"


class EnvironmentContext(SelfDescribingJson):
    """Environment context for the Snowplow tracker."""

    def __init__(self):
        """Initialize the environment context."""
        ci_markers = ("GITHUB_ACTIONS", "CI")
        super().__init__(
            EnvironmentContextSchema.url,
            {
                "context_uuid": str(uuid.uuid4()),
                "meltano_version": meltano.__version__,
                "is_dev_build": not release_marker_path.exists(),
                "is_ci_environment": any(
                    # True if 'true', 'TRUE', 'True', or '1'
                    os.environ.get(marker, "").lower()[:1] in {"1", "t"}
                    for marker in ci_markers
                ),
                "python_version": platform.python_version(),
                "python_implementation": platform.python_implementation(),
                **self.system_info,
                **self.process_info,
            },
        )

    @cached_property
    def system_info(self) -> dict[str, Any]:
        """Get system information.

        Returns:
            A dictionary containing system information.
        """
        try:
            freedesktop_data = platform.freedesktop_os_release()
        except Exception:
            freedesktop_data = defaultdict(type(None))

        return {
            "system_name": platform.system() or None,
            "system_release": platform.release() or None,
            "system_version": platform.version() or None,
            "machine": platform.machine() or None,
            "windows_edition": platform.win32_edition()
            if safe_hasattr(platform, "win32_edition")
            else None,
            "freedesktop_id": freedesktop_data["ID"],
            "freedesktop_id_like": freedesktop_data.get("ID_LIKE", None),
            "freedesktop_version_id": freedesktop_data.get("VERSION_ID", None),
        }

    @staticmethod
    def get_process_timestamp(process: psutil.Process) -> str:
        """Obtain the creation time of a process as a ISO 8601 timestamp.

        Args:
            process: The process to obtain the creation time from.

        Returns:
            A ISO 8601 timestamp formatted string.
        """
        return f"{datetime.utcfromtimestamp(process.create_time()).isoformat()}Z"

    @cached_property
    def process_info(self) -> dict[str, Any]:
        """Obtain the process information for the current process.

        Returns:
            A dictionary containing the process information. Such as the hashed process name, pid, core counts, etc
        """
        process = psutil.Process()
        with process.oneshot():
            return {
                "num_cpu_cores": psutil.cpu_count(),
                "num_cpu_cores_available": self.num_available_cores,
                "process_hierarchy": [
                    {
                        "process_name_hash": hash_sha256(proc.name()),
                        "process_creation_timestamp": self.get_process_timestamp(proc),
                    }
                    for proc in (process, *process.parents())
                ],
            }

    @cached_property
    def num_available_cores(self) -> int:
        """Obtain the number of available CPU cores.

        Uses sched_getaffinity where available, otherwise falls back to cpu_count().

        Returns:
            int: The number of available CPU cores.
        """
        if safe_hasattr(os, "sched_getaffinity"):
            return len(os.sched_getaffinity(0))
        return os.cpu_count()


environment_context = EnvironmentContext()



"""Module for handling multiple project .yml files."""

from __future__ import annotations

import copy
import logging
from collections import OrderedDict
from os import PathLike
from pathlib import Path
from typing import Any, Mapping, MutableMapping, Sequence, TypeVar

from atomicwrites import atomic_write
from ruamel.yaml import CommentedMap, CommentedSeq, YAMLError

from meltano.core import yaml

logger = logging.getLogger(__name__)
TMapping = TypeVar("TMapping", bound=MutableMapping)

BLANK_SUBFILE = CommentedMap(
    [
        ("plugins", {}),
        ("schedules", []),
        ("jobs", []),
        ("environments", []),
    ]
)
MULTI_FILE_KEYS = {
    "plugins",
    "schedules",
    "environments",
    "jobs",
}


def deep_merge(parent: TMapping, children: list[TMapping]) -> TMapping:
    """Deep merge a list of child dicts with a given parent.

    Args:
        parent: The parent dict.
        children: The child dicts.

    Returns:
        The merged dict.
    """
    base = copy.copy(parent)
    for child in children:
        for key, value in child.items():
            if isinstance(value, Mapping):
                # get node or create one
                node = base.setdefault(key, value.__class__())
                base[key] = deep_merge(node, [value])
            elif isinstance(value, Sequence):
                node = base.setdefault(key, value.__class__())
                node.extend(value)
            else:
                base[key] = value
    return base


class InvalidIncludePathError(Exception):
    """Occurs when an included file path matches a provided pattern but is not a valid config file."""


class ProjectFiles:  # noqa: WPS214
    """Interface for working with multiple project yaml files."""

    def __init__(self, root: Path, meltano_file_path: Path) -> None:
        """Instantiate ProjectFiles interface from project root and meltano.yml path.

        Args:
            root: The project root path.
            meltano_file_path: The path to the meltano.yml file.
        """
        self.root = root.resolve()
        self._meltano_file_path = meltano_file_path.resolve()
        self._plugin_file_map = {}
        self._raw_contents_map: dict[str, CommentedMap] = {}
        self._cached_loaded: CommentedMap | None = None

    @property
    def meltano(self) -> CommentedMap:
        """Return the contents of this project's `meltano.yml`.

        Returns:
            The contents of this projects `meltano.yml`.
        """
        return yaml.load(self._meltano_file_path)

    @property
    def include_paths(self) -> list[Path]:
        """Return list of paths derived from glob patterns defined in the meltanofile.

        Returns:
            List of paths derived from glob patterns defined in the meltanofile.
        """
        include_path_patterns = self.meltano.get("include_paths", [])
        return self._resolve_include_paths(include_path_patterns)

    def load(self) -> CommentedMap:
        """Load all project files into a single dict representation.

        Returns:
            A dict representation of all project files.
        """
        prev_raw_contents_map = self._raw_contents_map.copy()
        self._raw_contents_map.clear()
        self._raw_contents_map[str(self._meltano_file_path)] = self.meltano
        included_file_contents = self._load_included_files()

        # If the exact same objects are loaded again, use the cached result:
        k = TypeVar("k")

        def id_vals(x: dict[k, Any]) -> dict[k, int]:
            return {k: id(v) for k, v in x.items()}

        if self._cached_loaded is None or id_vals(prev_raw_contents_map) != id_vals(
            self._raw_contents_map
        ):
            self._cached_loaded = deep_merge(self.meltano, included_file_contents)

        return self._cached_loaded

    def update(self, meltano_config: dict) -> dict:
        """Update config by overriding current config with new, changed config.

        Note: `.update()` will write blank entities for those no longer in use
        (i.e. contained config on load, but not on save).

        Args:
            meltano_config: The new config to update with.

        Returns:
            The updated config dictionary.
        """
        file_dicts = self._split_config_dict(meltano_config)
        for file_path, contents in file_dicts.items():
            self._write_file(file_path, contents)

        unused_files = [fl for fl in self.include_paths if str(fl) not in file_dicts]
        for unused_file_path in unused_files:
            self._write_file(unused_file_path, BLANK_SUBFILE)
        return meltano_config

    def _is_valid_include_path(self, file_path: Path) -> None:
        """Determine if given path is a valid `include_paths` file.

        Args:
            file_path: The path to check.

        Raises:
            InvalidIncludePathError: If the included path is not a valid file.
        """
        if not (file_path.is_file() and file_path.exists()):
            raise InvalidIncludePathError(f"Included path '{file_path}' not found.")

    def _resolve_include_paths(self, include_path_patterns: list[str]) -> list[Path]:
        """Return a list of paths from a list of glob pattern strings.

        Not including `meltano.yml` (even if it is matched by a pattern).

        Args:
            include_path_patterns: List of glob pattern strings.

        Returns:
            List of paths matching the given glob patterns.

        Raises:
            InvalidIncludePathError: If a path is matched by a pattern but is not a valid
                file.
        """
        include_paths = []
        for pattern in include_path_patterns:
            for path in self.root.glob(pattern):
                try:
                    self._is_valid_include_path(path)
                except InvalidIncludePathError as err:
                    logger.critical(f"Include path '{path}' is invalid: \n {err}")
                    raise err
                include_paths.append(path)
            if self._meltano_file_path in include_paths:
                include_paths.remove(self._meltano_file_path)

        # Deduplicate entries
        return list(OrderedDict.fromkeys(include_paths))

    def _add_to_index(self, key: tuple, include_path: Path) -> None:
        """Add a new key:path to the `_plugin_file_map`.

        Args:
            key: The key to add.
            include_path: The path to add.

        Raises:
            Exception: If the plugin file is already in the index.
        """
        if key in self._plugin_file_map:
            key_path_string = ":".join(key)
            existing_key_file_path = self._plugin_file_map.get(key)
            logger.critical(
                f'Plugin with path "{key_path_string}" already added in file {existing_key_file_path}.'
            )
            raise Exception("Duplicate plugin name found.")
        else:
            self._plugin_file_map.update({key: str(include_path)})

    def _index_file(  # noqa: WPS210
        self, include_file_path: Path, include_file_contents: CommentedMap
    ) -> None:
        """Populate map of plugins/schedules to their respective files.

        This allows us to know exactly which plugin is configured where when we come to
        update plugins.

        Args:
            include_file_path: The path to the included file.
            include_file_contents: The contents of the included file.
        """
        # index plugins
        all_plugins = include_file_contents.get("plugins", {})
        for plugin_type, plugins in all_plugins.items():
            for plugin in plugins:
                plugin_key = ("plugins", plugin_type, plugin["name"])
                self._add_to_index(key=plugin_key, include_path=include_file_path)
        # index schedules
        schedules = include_file_contents.get("schedules", [])
        for schedule in schedules:
            schedule_key = ("schedules", schedule["name"])
            self._add_to_index(key=schedule_key, include_path=include_file_path)
        # index environments
        environments = include_file_contents.get("environments", [])
        for environment in environments:
            environment_key = ("environments", environment["name"])
            self._add_to_index(key=environment_key, include_path=include_file_path)

        jobs = include_file_contents.get("jobs", [])
        for job in jobs:
            job_key = ("jobs", job["name"])
            self._add_to_index(key=job_key, include_path=include_file_path)

    def _load_included_files(self) -> list[CommentedMap]:
        """Read and index included files.

        Returns:
            A list representation of all included files.

        Raises:
            YAMLError: If a file is invalid YAML.
        """
        self._plugin_file_map.clear()
        included_file_contents = []
        for path in self.include_paths:
            try:
                contents: CommentedMap = yaml.load(path)
            except YAMLError as exc:
                logger.critical(f"Error while parsing YAML file: {path} \n {exc}")
                raise exc
            else:
                self._raw_contents_map[str(path)] = contents
                # TODO: validate dict schema (https://gitlab.com/meltano/meltano/-/issues/3029)
                self._index_file(include_file_path=path, include_file_contents=contents)
                included_file_contents.append(contents)
        return included_file_contents

    def _add_mapping_entry(self, file_dicts, file, key, value):
        file_dict = file_dicts.setdefault(file, CommentedMap())
        entries = file_dict.setdefault(key, CommentedSeq())
        if value["name"] not in {scd["name"] for scd in entries}:
            entries.append(value)

    def _add_sequence_entry(self, file_dicts, key, elements):
        for elem in elements:
            file_key = (key, elem["name"])
            file = self._plugin_file_map.get(file_key, str(self._meltano_file_path))
            self._add_mapping_entry(file_dicts, file, key, elem)

    def _add_plugin(self, file_dicts, file, plugin_type, plugin):
        file_dict = file_dicts.setdefault(file, CommentedMap())
        plugins_dict = file_dict.setdefault("plugins", CommentedMap())
        plugins = plugins_dict.setdefault(plugin_type, CommentedSeq())
        if plugin["name"] not in {plg["name"] for plg in plugins}:
            plugins.append(plugin)

    def _add_plugins(self, file_dicts, all_plugins):
        for plugin_type, plugins in all_plugins.items():
            plugin_type = str(plugin_type)
            for plugin in plugins:
                key = ("plugins", plugin_type, plugin.get("name"))
                file = self._plugin_file_map.get(key, str(self._meltano_file_path))
                self._add_plugin(file_dicts, file, plugin_type, plugin)

    def _split_config_dict(self, config: CommentedMap):
        file_dicts: dict[str, CommentedMap] = {}

        # Fill in the top-level entries
        for key, value in config.items():
            if key == "plugins":
                self._add_plugins(file_dicts, value)
            elif key in MULTI_FILE_KEYS:
                self._add_sequence_entry(file_dicts, key, value)
            else:
                file = str(self._meltano_file_path)
                file_dict = file_dicts.setdefault(file, CommentedMap())
                file_dict[key] = value

        # Make sure that the top-level keys are in the same order as the original files
        sorted_file_dicts = self._restore_file_key_order(file_dicts)

        # Copy comments from the original config to the new config
        config.copy_attributes(sorted_file_dicts[str(self._meltano_file_path)])
        self._copy_yaml_attributes(sorted_file_dicts)
        return sorted_file_dicts

    def _restore_file_key_order(self, file_dicts: dict[str, CommentedMap]):
        """Restore the order of the keys in the meltano project files.

        Args:
            file_dicts: The dictionary mapping included files to their contents.

        Returns:
            The file contents mapping with order preserved from the original files.∫
        """
        sorted_file_dicts: dict[str, CommentedMap] = {}

        for file, contents in self._raw_contents_map.items():
            if file not in file_dicts:
                continue

            new_keys = [key for key in file_dicts[file] if key not in contents]

            # Restore sorting in project files
            sorted_file_dicts[file] = CommentedMap()
            for key in contents.keys():
                if key in file_dicts[file]:
                    sorted_file_dicts[file][key] = file_dicts[file][key]  # noqa: WPS529

            # Add the new keys at the end
            for new_key in new_keys:
                sorted_file_dicts[file][new_key] = file_dicts[file][new_key]

        return sorted_file_dicts

    def _copy_yaml_attributes(  # noqa: WPS210
        self,
        file_dicts: dict[str, CommentedMap],
    ):
        """Restore comments from top-level YAML entries in project files.

        For every file, use the top-level entries (e.g. "plugins") to copy the comments
        from the original config.

        Args:
            file_dicts: The dictionary mapping included files to their contents.
        """
        for file, file_dict in file_dicts.items():
            original_contents = self._raw_contents_map[file]
            original_contents.copy_attributes(file_dict)

            original_environments = original_contents.get(
                "environments", CommentedSeq()
            )
            environments = file_dict.get("environments", CommentedSeq())
            original_environments.copy_attributes(environments)

            original_jobs = original_contents.get("jobs", CommentedSeq())
            jobs = file_dict.get("jobs", CommentedSeq())
            original_jobs.copy_attributes(jobs)

            original_plugins = original_contents.get("plugins", CommentedMap())
            plugins = file_dict.get("plugins", CommentedMap())
            for plugin_type, plugin_type_plugins in plugins.items():
                original_plugin_type_plugins = original_plugins.get(
                    plugin_type, CommentedSeq()
                )
                original_plugin_type_plugins.copy_attributes(plugin_type_plugins)

            original_schedules = original_contents.get("schedules", CommentedSeq())
            schedules = file_dict.get("schedules", CommentedSeq())
            original_schedules.copy_attributes(schedules)

    def _write_file(self, file_path: PathLike, contents: Mapping):
        with atomic_write(file_path, overwrite=True) as fl:
            yaml.dump(contents, fl)



"""Definition of the top-level Click group for the Meltano CLI."""

from __future__ import annotations

import logging
import sys
from typing import NoReturn

import click

import meltano
from meltano.cli.utils import InstrumentedGroup
from meltano.core.behavior.versioned import IncompatibleVersionError
from meltano.core.logging import LEVELS, setup_logging
from meltano.core.project import Project, ProjectNotFound
from meltano.core.project_settings_service import ProjectSettingsService
from meltano.core.tracking import CliContext, Tracker

logger = logging.getLogger(__name__)


class NoWindowsGlobbingGroup(InstrumentedGroup):
    """A instrumented Click group that does not perform glob expansion on Windows.

    This restores the behaviour of Click's globbing to how it was before v8.
    Click (as of version 8.1.3) ignores quotes around an asterisk, which makes
    it behave differently than most shells that support globbing, and make some
    typical Meltano commands fail, e.g. `meltano select tap-gitlab tags "*"`.
    """

    def main(self, *args, **kwargs) -> NoReturn:
        """Invoke the Click CLI with Windows globbing disabled.

        Args:
            args: Positional arguments for the Click group.
            kwargs: Keyword arguments for the Click group.
        """
        return super().main(*args, windows_expand_args=False, **kwargs)


@click.group(
    cls=NoWindowsGlobbingGroup, invoke_without_command=True, no_args_is_help=True
)
@click.option("--log-level", type=click.Choice(LEVELS.keys()))
@click.option(
    "--log-config", type=str, help="Path to a python logging yaml config file."
)
@click.option("-v", "--verbose", count=True, help="Not used.")
@click.option(
    "--environment",
    envvar="MELTANO_ENVIRONMENT",
    help="Meltano environment name.",
)
@click.option(
    "--no-environment", is_flag=True, default=False, help="Don't use any environment."
)
@click.version_option(version=meltano.__version__, prog_name="meltano")
@click.pass_context
def cli(  # noqa: WPS231
    ctx,
    log_level: str,
    log_config: str,
    verbose: int,
    environment: str,
    no_environment: bool,
):  # noqa: WPS231
    """
    ELT for the DataOps era.

    \b\nRead more at https://docs.meltano.com/reference/command-line-interface
    """
    ctx.ensure_object(dict)

    if log_level:
        ProjectSettingsService.config_override["cli.log_level"] = log_level

    if log_config:
        ProjectSettingsService.config_override["cli.log_config"] = log_config

    ctx.obj["verbosity"] = verbose

    try:  # noqa: WPS229
        project = Project.find()
        setup_logging(project)
        project_setting_service = ProjectSettingsService(project)

        readonly = project_setting_service.get("project_readonly")
        if readonly:
            project.readonly = True
        if project.readonly:
            logger.debug("Project is read-only.")

        # detect active environment
        selected_environment = None
        is_default_environment = False
        if no_environment or (environment and environment.lower() == "null"):
            logger.info("No environment is active")
        elif environment:
            selected_environment = environment
        elif project_setting_service.get("default_environment"):
            selected_environment = project_setting_service.get("default_environment")
            is_default_environment = True
        # activate environment
        if selected_environment:
            project.activate_environment(selected_environment)
            logger.info(
                "Environment '%s' is active", selected_environment  # noqa: WPS323
            )
        ctx.obj["is_default_environment"] = is_default_environment
        ctx.obj["project"] = project
        ctx.obj["tracker"] = Tracker(project)
        ctx.obj["tracker"].add_contexts(
            CliContext.from_click_context(ctx)
        )  # backfill the `cli` CliContext
    except ProjectNotFound:
        ctx.obj["project"] = None
    except IncompatibleVersionError:
        click.secho(
            "This Meltano project is incompatible with this version of `meltano`.",
            fg="yellow",
        )
        click.echo(
            "For more details, visit https://docs.meltano.com/guide/installation#upgrading-meltano-version"
        )
        sys.exit(3)
